import re

__all__ = ['query_key_handler']

versioning_regex = re.compile(r'\.([a-z])', re.IGNORECASE)


def query_key_handler(query_id, request):
    if request.version is None:
        return query_id

    query_prefix = versioning_regex.sub(r':\1', request.version)
    return ':'.join((query_prefix, query_id))



from django.core.cache import cache as default_cache
from django.core.cache import caches
from django.core.cache.backends.base import InvalidCacheBackendError

from .settings import persist_settings

try:
    cache = caches[persist_settings.CACHE_NAME]
except (InvalidCacheBackendError, ValueError):
    cache = default_cache



from collections import OrderedDict

__all__ = [
    'BaseRenderer',
    'BaseStripTagsRenderer',
]


def strip_tags(data, f):
    data = f(data)
    if isinstance(data, dict):
        return OrderedDict(
            (k, strip_tags(v, f))
            for k, v in data.items())
    elif isinstance(data, list):
        return [strip_tags(v, f) for v in data]
    return data


class BaseRenderer:

    def render(self, data, context=None):
        raise NotImplementedError('.render() must be implemented')


class BaseStripTagsRenderer:

    def render(self, data, context=None):
        assert hasattr(self, 'strip_func'), (
            '`{cls}.strip_func` argument is required'
            .format(cls=self.__class__.__name__)
        )
        return strip_tags(data, self.strip_func)



import os

from django.utils._os import safe_join

from .base import BaseLoader, Origin
from .exceptions import DocumentDoesNotExist


class FilesystemLoader(BaseLoader):

    def get_dirs(self):
        return self.engine.dirs

    def get_sources(self, query_key):
        document_name = query_key + self.engine.documents_ext
        query_keys = document_name.split(':')

        for document_dir in self.get_dirs():
            for s in range(len(query_keys)):
                origin_keys = query_keys[:-s - 1] + query_keys[-1:]
                origin_path = safe_join(document_dir, *origin_keys)

                if os.path.isfile(origin_path):
                    yield Origin(
                        name=origin_path,
                        query_key=query_key,
                        loader=self)

    def get_contents(self, origin):
        try:
            with open(origin.name, 'r') as document:
                return document.read()
        except FileNotFoundError:
            raise DocumentDoesNotExist(origin)



from ..settings import persist_settings
from .exceptions import DocumentDoesNotExist


class CachedEngine:
    dirs = persist_settings.DOCUMENTS_DIRS
    documents_ext = persist_settings.DOCUMENTS_EXT

    def __init__(self, dirs=None):
        self.get_document_cache = {}
        self.loaders = self.get_loaders()

    def get_document(self, query_id, request):
        query_key = persist_settings.QUERY_KEY_HANDLER(query_id, request)
        document = self.get_document_cache.get(query_key)

        if document is None:
            document = self.find_document(query_key)
            self.get_document_cache[query_key] = document
        return document

    def find_document(self, query_key):
        for loader in self.loaders:
            try:
                return loader.get_document(query_key)
            except DocumentDoesNotExist:
                continue
        raise DocumentDoesNotExist(query_key)

    def get_loaders(self):
        loader_classes = persist_settings.DEFAULT_LOADER_CLASSES
        return [loader_class(self) for loader_class in loader_classes]

    def get_sources(self, query_key):
        for loader in self.loaders:
            yield from loader.get_sources(query_key)

    def reset(self):
        self.get_document_cache.clear()



import os

from django.core.exceptions import ValidationError
from django.core.validators import URLValidator

import requests

from .base import BaseLoader, Origin

__all__ = ['URLLoader']


class URLOrigin(Origin):

    def __init__(self, content, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.content = content


class URLLoader(BaseLoader):

    def get_dirs(self):
        return self.engine.dirs

    def get_sources(self, query_key):
        document_name = query_key + self.engine.documents_ext
        query_keys = document_name.split(':')

        for document_dir in self.get_dirs():
            origin_url = os.path.join(document_dir, *query_keys)

            try:
                URLValidator()(origin_url)
            except ValidationError:
                continue
            try:
                response = requests.get(origin_url)
            except requests.HTTPError:
                continue

            if response.status_code == 200:
                yield URLOrigin(
                    content=response.text,
                    name=origin_url,
                    query_key=query_key,
                    loader=self)

    def get_contents(self, origin):
        return origin.content



from collections import OrderedDict

from .base import BaseStripTagsRenderer

__all__ = ['StripRelayTagsRenderer']


def strip_edges(data):
    if isinstance(data, dict):
        edges = data.get('edges')
        if edges is not None:
            if len(data) == 1:
                return edges
            return OrderedDict(
                ('results', edges) if k == 'edges'
                else (k, v) for k, v in data.items())
        elif 'node' in data:
            return data['node']
    return data


class StripRelayTagsRenderer(BaseStripTagsRenderer):
    strip_func = staticmethod(strip_edges)



import re

from django.http.multipartparser import parse_header
from django.utils.translation import ugettext_lazy as _

from . import exceptions
from .settings import persist_settings


class BaseVersioning:
    default_version = persist_settings.DEFAULT_VERSION
    allowed_versions = persist_settings.ALLOWED_VERSIONS

    def get_version(self, request):
        raise NotImplementedError('.get_version() must be implemented')

    def is_allowed_version(self, version):
        if not self.allowed_versions:
            return True
        return (version == self.default_version or
                version in self.allowed_versions)


class AcceptHeaderVersioning(BaseVersioning):
    version_param = persist_settings.VERSION_PARAM

    def parse_header(self, media_type):
        base_media_type, params = parse_header(media_type)
        return params.get(self.version_param, self.default_version)

    def get_version(self, request):
        media_type = request.META.get('HTTP_ACCEPT', b'').encode()
        version = self.parse_header(media_type)

        if hasattr(version, 'decode'):
            version = version.decode('iso-8859-1')

        if not self.is_allowed_version(version):
            raise exceptions.GraphQLPersistError(
                _('Invalid version in "Accept" header'))
        return version


class VendorTreeVersioning(AcceptHeaderVersioning):
    media_type_regex = re.compile(
        r'^application/vnd\.{}\.([a-z0-9\.]+[a-z0-9]+)\+json$'
        .format(persist_settings.MEDIA_TYPE_NAME).encode(),
        re.IGNORECASE)

    def parse_header(self, media_type):
        match = self.media_type_regex.match(media_type)

        if not match:
            return self.default_version
        return match.group(1)


class QueryParameterVersioning(BaseVersioning):
    version_param = persist_settings.VERSION_PARAM

    def get_version(self, request):
        version = request.GET.get(self.version_param, self.default_version)

        if not self.is_allowed_version(version):
            raise exceptions.GraphQLPersistError(
                _('Invalid version in query parameter'))
        return version


class HostNameVersioning(BaseVersioning):
    hostname_regex = re.compile(
        r'^([a-z0-9]+)\.[a-z0-9]+\.[a-z0-9]+$',
        re.IGNORECASE)

    def get_version(self, request):
        hostname, separator, port = request.get_host().partition(':')
        match = self.hostname_regex.match(hostname)

        if not match:
            return self.default_version

        version = match.group(1)

        if not self.is_allowed_version(version):
            raise exceptions.GraphQLPersistError(
                _('Invalid version in hostname'))
        return version



from django.template.utils import get_app_template_dirs

from ..settings import persist_settings
from .filesystem import FilesystemLoader


class AppDirectoriesLoader(FilesystemLoader):

    def get_dirs(self):
        return get_app_template_dirs(persist_settings.APP_DOCUMENT_DIR)



from .app_directories import AppDirectoriesLoader
from .base import BaseLoader
from .cached import CachedEngine
from .filesystem import FilesystemLoader
from .url import URLLoader

__all__ = [
    'AppDirectoriesLoader',
    'BaseLoader',
    'CachedEngine',
    'FilesystemLoader',
    'URLLoader',
]



from .exceptions import DocumentDoesNotExist


class Origin:

    def __init__(self, name, query_key=None, loader=None):
        self.name = name
        self.query_key = query_key
        self.loader = loader

    def __str__(self):
        return self.name


class Document:

    def __init__(self, source, origin=None):
        self.source = source
        self.origin = origin

    def parse(self):
        return self.source


class BaseLoader:

    def __init__(self, engine, dirs=None):
        self.engine = engine

    def get_document(self, query_key):
        for origin in self.get_sources(query_key):
            try:
                contents = self.get_contents(origin)
            except DocumentDoesNotExist:
                continue
            return Document(source=contents, origin=origin)
        raise DocumentDoesNotExist(query_key)

    def get_sources(self, query_key):
        raise NotImplementedError('.get_sources() must be implemented')

    def get_contents(self, origin):
        raise NotImplementedError('.get_contents() must be implemented')



from importlib import import_module

from django.conf import settings
from django.test.signals import setting_changed

DEFAULTS = {
    'DOCUMENTS_DIRS': (),
    'DOCUMENTS_EXT': '.graphql',
    'CACHE_NAME': 'default',
    'QUERY_KEY_HANDLER': 'graphql_persist.query.query_key_handler',
    'DEFAULT_VERSIONING_CLASS': None,
    'DEFAULT_LOADER_CLASSES': (
        'graphql_persist.loaders.AppDirectoriesLoader',
        'graphql_persist.loaders.FilesystemLoader',
        'graphql_persist.loaders.URLLoader',
    ),
    'DEFAULT_RENDERER_CLASSES': (),

    # Versioning
    'DEFAULT_VERSION': None,
    'ALLOWED_VERSIONS': None,
    'VERSION_PARAM': 'version',
    'MEDIA_TYPE_NAME': r'[a-zA-Z0-9]+',

    # Loaders
    'APP_DOCUMENT_DIR': 'documents',
}

IMPORT_STRINGS = (
    'QUERY_KEY_HANDLER',
    'DEFAULT_VERSIONING_CLASS',
    'DEFAULT_LOADER_CLASSES',
    'DEFAULT_RENDERER_CLASSES',
)


def perform_import(value, setting_name):
    if value is None:
        return None
    if isinstance(value, str):
        return import_from_string(value, setting_name)
    elif isinstance(value, (list, tuple)):
        return [import_from_string(item, setting_name) for item in value]
    return value


def import_from_string(value, setting_name):
    try:
        module_path, class_name = value.rsplit('.', 1)
        module = import_module(module_path)
        return getattr(module, class_name)
    except (ImportError, AttributeError) as e:
        msg = 'Could not import `{}` for Persist setting `{}`. {}: {}.'.format(
            value, setting_name, e.__class__.__name__, e)
        raise ImportError(msg)


class PersistSettings:

    def __init__(self, defaults, import_strings):
        self.defaults = defaults
        self.import_strings = import_strings
        self._cached_attrs = set()

    def __getattr__(self, attr):
        if attr not in self.defaults:
            raise AttributeError('Invalid setting: `{}`'.format(attr))

        value = self.user_settings.get(attr, self.defaults[attr])

        if attr in self.import_strings:
            value = perform_import(value, attr)

        self._cached_attrs.add(attr)
        setattr(self, attr, value)
        return value

    @property
    def user_settings(self):
        if not hasattr(self, '_user_settings'):
            self._user_settings = getattr(settings, 'GRAPHQL_PERSIST', {})
        return self._user_settings

    def reload(self):
        for attr in self._cached_attrs:
            delattr(self, attr)

        self._cached_attrs.clear()

        if hasattr(self, '_user_settings'):
            delattr(self, '_user_settings')


def reload_settings(*args, **kwargs):
    setting = kwargs['setting']

    if setting == 'GRAPHQL_PERSIST':
        persist_settings.reload()


setting_changed.connect(reload_settings)

persist_settings = PersistSettings(DEFAULTS, IMPORT_STRINGS)



import json
from collections import OrderedDict

from django.utils.encoding import force_text

from graphene_django.views import GraphQLView

from . import exceptions
from .loaders import CachedEngine
from .loaders.exceptions import DocumentDoesNotExist, DocumentSyntaxError
from .settings import persist_settings

__all__ = ['PersistMiddleware']


def get_json_data(content, **kwargs):
    content = force_text(content, **kwargs)
    return json.loads(content, object_pairs_hook=OrderedDict)


class PersistMiddleware:

    def __init__(self, get_response):
        self.get_response = get_response
        self.loader = CachedEngine()
        self.renderers = self.get_renderers()
        self.versioning_class = persist_settings.DEFAULT_VERSIONING_CLASS

    def __call__(self, request):
        try:
            request.version = self.get_version(request)
        except exceptions.GraphQLPersistError as err:
            return exceptions.PersistResponseError(str(err))

        response = self.get_response(request)

        if (response.status_code == 200 and
                hasattr(request, 'persisted_query') and
                self.renderers):

            self.finalize_response(request, response)
        return response

    def process_view(self, request, view_func, *args):
        if (hasattr(view_func, 'view_class') and
                issubclass(view_func.view_class, GraphQLView) and
                request.content_type == 'application/json'):

            try:
                data = get_json_data(request.body)
            except json.JSONDecodeError:
                return None

            query_id = data.get('id', data.get('operationName'))

            if query_id and not data.get('query'):
                try:
                    document = self.loader.get_document(query_id, request)
                except DocumentDoesNotExist as err:
                    return exceptions.DocumentNotFound(str(err))

                try:
                    data['query'] = document.parse()
                except DocumentSyntaxError as err:
                    return exceptions.DocumentSyntaxError(str(err))

                request.persisted_query = document
                request._body = json.dumps(data).encode()
        return None

    def get_version(self, request):
        if self.versioning_class is not None:
            return self.versioning_class().get_version(request)
        return None

    def get_renderers(self):
        renderer_classes = persist_settings.DEFAULT_RENDERER_CLASSES
        return [renderer() for renderer in renderer_classes]

    def finalize_response(self, request, response):
        data = get_json_data(response.content, encoding=response.charset)

        context = {
            'request': request,
        }

        for renderer in self.renderers:
            data = renderer.render(data, context)

        response.content = json.dumps(data)

        if response.has_header('Content-Length'):
            response['Content-Length'] = str(len(response.content))




class DocumentDoesNotExist(Exception):
    """Document Does Not Exist"""


class DocumentSyntaxError(Exception):

    def __init__(self, query_key, syntax_error=None):
        self.syntax_error = syntax_error
        return super().__init__(query_key)



from django.http import JsonResponse
from django.utils.translation import ugettext_lazy as _


class GraphQLPersistError(Exception):
    """GraphQL Persist Error"""


class PersistResponseError(JsonResponse):
    default_message = _('A server error occurred')
    status_code = 400

    def __init__(self, message=None):
        if message is None:
            message = self.default_message

        return super().__init__({
            'errors': [{'message': message}],
        })


class DocumentNotFound(PersistResponseError):
    status_code = 404

    def __init__(self, query_key):
        message = _('Document `{}` not found').format(query_key)
        return super().__init__(message)


class DocumentSyntaxError(PersistResponseError):

    def __init__(self, query_key):
        message = _('Syntax Error on `{}`').format(query_key)
        return super().__init__(message)
